# Generated by Django 4.0 on 2023-10-13 20:19 by Hesham

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_basin3d', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attributemapping',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='datasource',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
